/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wFluid.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
using namespace tbox;
using namespace dart;

Fluid::Fluid(std::shared_ptr<MshData> _msh, int no,
             double mInf, int dim, double alpha, double beta) : Group(_msh, no)
{
    initFun(mInf, dim, alpha, beta);
    createMap();
    std::cout << "Fluid is " << *tag << "with " << nCnt << " nodes."
              << std::endl
              << std::endl;
}
Fluid::Fluid(std::shared_ptr<MshData> _msh, std::string const &name,
             double mInf, int dim, double alpha, double beta) : Group(_msh, name)
{
    initFun(mInf, dim, alpha, beta);
    createMap();
    std::cout << "Fluid is " << *tag << "with " << nCnt << " nodes."
              << std::endl
              << std::endl;
}
Fluid::~Fluid()
{
    delete rho;
    delete mach;
    delete cP;
    std::cout << "~Fluid()\n";
}

/**
 * Initialize functions
 */
void Fluid::initFun(double mInf, int dim, double alpha, double beta)
{
    if (mInf == 0)
    {
        rho = new F0ElRhoLin();
        mach = new F0ElMachLin();
        cP = new F0ElCpLin();
    }
    else
    {
        rho = new F0ElRhoClamp(mInf, 3.0);
        mach = new F0ElMachClamp(mInf, 3.0);
        cP = new F0ElCpClamp(mInf, 3.0);
    }
    phiInf = new F0PsPhiInf(dim, alpha, beta);
}

/**
 * @brief Create data structure
 */
void Fluid::createMap()
{
    // Create map of elements connected to node
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            neMap[n].push_back(e);
    nCnt = neMap.size();
    // Collect connected elements
    adjMap.resize(tag->elems.size());
    int i = 0;
    for (auto e : tag->elems)
    {
        // Get elements sharing at least one node
        std::vector<Element *> vAdjE;
        for (auto n : e->nodes)
            vAdjE.insert(vAdjE.end(), neMap.at(n).begin(), neMap.at(n).end());
        // Sort and remove duplicates
        std::sort(vAdjE.begin(), vAdjE.end());
        auto iterator = std::unique(vAdjE.begin(), vAdjE.end());
        vAdjE.resize(std::distance(vAdjE.begin(), iterator));
        // Remove self
        for (size_t j = 0; j < vAdjE.size(); ++j)
        {
            if (vAdjE[j] == e)
                vAdjE.erase(vAdjE.begin() + j);
        }
        // Pair an element and its adjacent elements
        adjMap[i] = std::pair<Element *, std::vector<Element *>>(e, vAdjE);
        adjMap2[e] = i;
        i++;
    }
}

void Fluid::write(std::ostream &out) const
{
    out << "Fluid on " << *tag << " with:\n\t"
        << "rho = " << this->rho << ", mach = " << this->mach << ", cP = " << this->cP << "\n";
    out << "Adjencency map:\n";
    for (auto p : adjMap)
    {
        out << "\t" << p.first->no << ':';
        for (size_t j = 0; j < p.second.size(); j++)
            out << p.second[j]->no << ',';
        out << ";\n";
    }
}
