/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wPotentialResidual.h"
#include "wFluid.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace dart;

/**
 * @brief Build the residual matrix for a fixed-point iteration, on one element
 *
 * A = \int rho * grad_phi * grad_psi dV
 */
Eigen::MatrixXd PotentialResidual::buildFixed(Element const &e, std::vector<double> const &phi, Fluid const &fluid)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Build
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions gradient
        Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
        // Elementary stiffness matrix
        A += dSf.transpose() * dSf * fluid.rho->eval(e, phi, k) * gauss.getW(k) * e.getDetJ(k);
    }
    return A;
}

/**
 * @brief Build the residual vector, on one element
 *
 * b = \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV
 */
Eigen::VectorXd PotentialResidual::build(Element const &e, Element const &eU, Element const &eUU, std::vector<double> const &phi, Fluid const &fluid, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::VectorXd b = Eigen::VectorXd::Zero(e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
        b += fluid.rho->eval(e, phi, k) * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);

    // Supersonic contribution
    double machC = fluid.mach->eval(e, phi, 0);
    double machU = fluid.mach->eval(eU, phi, 0);
    double mach = (machC < machU) ? machU : machC;
    if (mach > mCO)
    {
        double mu = muC * (1 - (mCO * mCO) / (mach * mach)); // switching  function
        double rhoU = fluid.rho->eval(eU, phi, 0);           // density on upwind element
        // scale 1st term
        b *= 1 - mu;
        // contribution of upwind element
        for (size_t k = 0; k < gauss.getN(); ++k)
            b += mu * rhoU * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);

        // second order upwind
        double rhoUU = fluid.rho->eval(eUU, phi, 0);
        double ratio = (e.computeGradient(phi, 0)(0) - eU.computeGradient(phi, 0)(0)) / (eU.computeGradient(phi, 0)(0) - eUU.computeGradient(phi, 0)(0));
        if (ratio >= 0.)
        {
            double alpha = 1.0;
            if (ratio >= 1.00)
                alpha = 1.00;
            else if (ratio >= 0.95)
                alpha = 0.95;
            else if (ratio >= 0.90)
                alpha = 0.90;
            else if (ratio >= 0.85)
                alpha = 0.85;
            else if (ratio >= 0.80)
                alpha = 0.80;
            else if (ratio >= 0.75)
                alpha = 0.75;
            else if (ratio >= 0.70)
                alpha = 0.70;
            else if (ratio >= 0.65)
                alpha = 0.65;
            else if (ratio >= 0.60)
                alpha = 0.60;
            else if (ratio >= 0.55)
                alpha = 0.55;
            else if (ratio >= 0.50)
                alpha = 0.50;
            else if (ratio >= 0.45)
                alpha = 0.45;
            else if (ratio >= 0.40)
                alpha = 0.40;
            else if (ratio >= 0.35)
                alpha = 0.35;
            else if (ratio >= 0.30)
                alpha = 0.30;
            else if (ratio >= 0.25)
                alpha = 0.25;
            else if (ratio >= 0.20)
                alpha = 0.20;
            else if (ratio >= 0.15)
                alpha = 0.15;
            else if (ratio >= 0.10)
                alpha = 0.10;
            else if (ratio >= 0.05)
                alpha = 0.05;
            else
                alpha = 0.025;
            //alpha = (alpha < 0.5) ? alpha*2 : 1.0;

            for (size_t k = 0; k < gauss.getN(); ++k)
            {
                b += alpha * mu * rhoU * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);
                b -= alpha * mu * rhoUU * (e.getJinv(k) * cache.getDsf(k)).transpose() * e.computeGradient(phi, k) * gauss.getW(k) * e.getDetJ(k);
            }
        }
    }
    return b;
}

/**
 * @brief Build the gradient of the residual vector with respect to the flow variable (jacobian), on one element
 *
 * A = \int d( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV
 *   = \int (1-mu)*drho * grad_phi * grad_psi dV
 *   + \int mu*drhoU * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * dgrad_phi * grad_psi dV
 *   + \int (rhoU - rho)*dmu * grad_phi * grad_psi dV
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd, Eigen::MatrixXd> PotentialResidual::buildGradientFlow(Element const &e, Element const &eU, Element const &eUU, std::vector<double> const &phi, Fluid const &fluid, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::MatrixXd A1 = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Gauss point and determinant
        double wdj = gauss.getW(k) * e.getDetJ(k);
        // Shape functions and solution gradient
        Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
        Eigen::VectorXd dPhi = e.computeGradient(phi, k);

        // rho * grad_phi*grad_psi + drho * grad_phi*grad_psi
        A1 += fluid.rho->eval(e, phi, k) * dSf.transpose() * dSf * wdj;
        A1 += fluid.rho->evalGrad(e, phi, k) * dSf.transpose() * dPhi * dPhi.transpose() * dSf * wdj;
    }

    // Supersonic contribution
    Eigen::MatrixXd A2, A3;
    double machC = fluid.mach->eval(e, phi, 0);
    double machU = fluid.mach->eval(eU, phi, 0);
    double mach = (machC < machU) ? machU : machC;
    if (mach > mCO)
    {
        // switching function and gradient
        double mu = muC * (1 - (mCO * mCO) / (mach * mach));
        double dmu = 2 * muC * mCO * mCO / (mach * mach * mach);
        // density and gradients on upwind elements
        Eigen::VectorXd dPhiU = eU.computeGradient(phi, 0);
        double rhoU = fluid.rho->eval(eU, phi, 0);
        double dRhoU = fluid.rho->evalGrad(eU, phi, 0);
        Eigen::MatrixXd const &dSfU = eU.getJinv(0) * eU.getVCache().getDsf(0);

        // upwinding
        A1 *= 1 - mu;
        A2 = Eigen::MatrixXd::Zero(e.nodes.size(), eU.nodes.size());
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Gauss point and determinant
            double wdj = gauss.getW(k) * e.getDetJ(k);
            // Shape functions and solution gradient
            Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
            Eigen::VectorXd dPhi = e.computeGradient(phi, k);

            // mu * drhoU * grad_phi * grad_psi
            A2 += mu * dRhoU * dSf.transpose() * dPhi * dPhiU.transpose() * dSfU * wdj;
            // mu * rhoU * grad_phi * grad_psi
            A1 += mu * rhoU * dSf.transpose() * dSf * wdj;
            // dmu * (rhoU-rho) * grad_phi * grad_psi OR dmuU * (rhoU-rho) * grad_phi * grad_psi
            if (machC >= machU)
                A1 += dmu * (rhoU - fluid.rho->eval(e, phi, k)) * fluid.mach->evalGrad(e, phi, k) * dSf.transpose() * dPhi * dPhi.transpose() * dSf * wdj;
            else
                A2 += dmu * (rhoU - fluid.rho->eval(e, phi, k)) * fluid.mach->evalGrad(eU, phi, k) * dSf.transpose() * dPhi * dPhiU.transpose() * dSfU * wdj;
        }

        // second order upwind
        double rhoUU = fluid.rho->eval(eUU, phi, 0);
        double ratio = (e.computeGradient(phi, 0)(0) - eU.computeGradient(phi, 0)(0)) / (eU.computeGradient(phi, 0)(0) - eUU.computeGradient(phi, 0)(0));
        if (ratio >= 0.)
        {
            double alpha = 1.0;
            if (ratio >= 1.0)
                alpha = 1.00;
            else if (ratio >= 0.95)
                alpha = 0.95;
            else if (ratio >= 0.90)
                alpha = 0.90;
            else if (ratio >= 0.85)
                alpha = 0.85;
            else if (ratio >= 0.80)
                alpha = 0.80;
            else if (ratio >= 0.75)
                alpha = 0.75;
            else if (ratio >= 0.70)
                alpha = 0.70;
            else if (ratio >= 0.65)
                alpha = 0.65;
            else if (ratio >= 0.60)
                alpha = 0.60;
            else if (ratio >= 0.55)
                alpha = 0.55;
            else if (ratio >= 0.50)
                alpha = 0.50;
            else if (ratio >= 0.45)
                alpha = 0.45;
            else if (ratio >= 0.40)
                alpha = 0.40;
            else if (ratio >= 0.35)
                alpha = 0.35;
            else if (ratio >= 0.30)
                alpha = 0.30;
            else if (ratio >= 0.25)
                alpha = 0.25;
            else if (ratio >= 0.20)
                alpha = 0.20;
            else if (ratio >= 0.15)
                alpha = 0.15;
            else if (ratio >= 0.10)
                alpha = 0.10;
            else if (ratio >= 0.05)
                alpha = 0.05;
            else
                alpha = 0.025;
            //alpha = (alpha < 0.5) ? alpha*2 : 1.0;

            double dRhoUU = fluid.rho->evalGrad(eUU, phi, 0);
            Eigen::VectorXd dPhiUU = eUU.computeGradient(phi, 0);
            Eigen::MatrixXd const &dSfUU = eUU.getJinv(0) * eUU.getVCache().getDsf(0);

            A3 = Eigen::MatrixXd::Zero(e.nodes.size(), eUU.nodes.size());
            for (size_t k = 0; k < gauss.getN(); ++k)
            {
                // Gauss point and determinant
                double wdj = gauss.getW(k) * e.getDetJ(k);
                // Shape functions and solution gradient
                Eigen::MatrixXd const &dSf = e.getJinv(k) * cache.getDsf(k);
                Eigen::VectorXd dPhi = e.computeGradient(phi, k);

                //  (mu * rhoU - mu * rhUU) * grad_phi * grad_psi
                A1 += alpha * mu * rhoU * dSf.transpose() * dSf * wdj;
                A1 -= alpha * mu * rhoUU * dSf.transpose() * dSf * wdj;
                // mu * drhoU * grad_phi * grad_psi
                A2 += alpha * mu * dRhoU * dSf.transpose() * dPhi * dPhiU.transpose() * dSfU * wdj;
                // mu * drhoUU * grad_phi * grad_psi
                A3 -= alpha * mu * dRhoUU * dSf.transpose() * dPhi * dPhiUU.transpose() * dSfUU * wdj;
                // dmu(U) * (rhoU-rhoUU) * grad_phi * grad_psi
                if (machC >= machU)
                    A1 += alpha * dmu * (rhoU - rhoUU) * fluid.mach->evalGrad(e, phi, k) * dSf.transpose() * dPhi * dPhi.transpose() * dSf * wdj;
                else
                    A2 += alpha * dmu * (rhoU - rhoUU) * fluid.mach->evalGrad(eU, phi, k) * dSf.transpose() * dPhi * dPhiU.transpose() * dSfU * wdj;
            }
        }
    }
    return std::make_tuple(A1, A2, A3);
}

/**
 * @brief Build the gradient of the residual vector with respect to the nodes, on one element
 *
 * A = d( \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi dV )
 *   = \int (1-mu)*drho * grad_phi * grad_psi dV
 *   + \int mu*drhoU * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * dgrad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * dgrad_psi dV
 *   + \int (rhoU - rho)*dmu * grad_phi * grad_psi dV
 *   + \int ( (1-mu)*rho + mu*rhoU ) * grad_phi * grad_psi ddV
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> PotentialResidual::buildGradientMesh(tbox::Element const &e, tbox::Element const &eU, std::vector<double> const &phi, Fluid const &fluid, int nDim, double muC, double mCO)
{
    // Get pre-computed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Subsonic contribution
    Eigen::MatrixXd A1 = Eigen::MatrixXd::Zero(e.nodes.size(), nDim * e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Density and gradients
        Eigen::VectorXd dPhi = e.computeGradient(phi, k);
        double rho = fluid.rho->eval(e, phi, k);
        double dRho = fluid.rho->evalGrad(e, phi, k);
        Eigen::MatrixXd const &dSf = cache.getDsf(k);
        Eigen::MatrixXd const &iJ = e.getJinv(k);
        // Jacobian gradients
        std::vector<Eigen::MatrixXd> const &dJ = e.getGradJ(k);
        std::vector<double> const &dDetJ = e.getGradDetJ(k);
        // Gauss points and determinant
        double w = gauss.getW(k);
        double detJ = e.getDetJ(k);
        for (size_t i = 0; i < e.nodes.size(); ++i)
        {
            for (int m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                A1.col(idx) += w * dRho * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * (iJ * dSf).transpose() * dPhi * detJ; // drho * grad_phi * grad_psi * detj
                A1.col(idx) += w * rho * (iJ * dSf).transpose() * (-iJ * dJ[idx] * dPhi) * detJ;                            // rho * dgrad_phi * grad_psi * detj
                A1.col(idx) += w * rho * (-iJ * dJ[idx] * iJ * dSf).transpose() * dPhi * detJ;                              // rho * grad_phi * dgrad_psi * detj
                A1.col(idx) += w * rho * (iJ * dSf).transpose() * dPhi * dDetJ[idx];                                        // rho * grad_phi * grad_psi * ddetj
            }
        }
    }
    // Supersonic contribution
    Eigen::MatrixXd A2;
    double machC = fluid.mach->eval(e, phi, 0);
    double machU = fluid.mach->eval(eU, phi, 0);
    double mach = (machC < machU) ? machU : machC;
    if (mach > mCO)
    {
        // switching function and gradient
        double mu = muC * (1 - (mCO * mCO) / (mach * mach));
        double dmu = 2 * muC * mCO * mCO / (mach * mach * mach);
        // upwind density and gradient
        Eigen::VectorXd dPhiU = eU.computeGradient(phi, 0);
        double rhoU = fluid.rho->eval(eU, phi, 0);
        double dRhoU = fluid.rho->evalGrad(eU, phi, 0);
        Eigen::MatrixXd const &iJU = eU.getJinv(0);
        std::vector<Eigen::MatrixXd> const &dJU = eU.getGradJ(0);

        // upwinding
        A1 *= 1 - mu;
        A2 = Eigen::MatrixXd::Zero(e.nodes.size(), nDim * eU.nodes.size());
        for (size_t k = 0; k < gauss.getN(); ++k)
        {
            // Density, Mach and Gradients
            Eigen::VectorXd dPhi = e.computeGradient(phi, k);
            double rho = fluid.rho->eval(e, phi, k);
            Eigen::MatrixXd const &dSf = cache.getDsf(k);
            Eigen::MatrixXd const &iJ = e.getJinv(k);
            // Jacobian gradients
            std::vector<Eigen::MatrixXd> const &dJ = e.getGradJ(k);
            std::vector<double> const &dDetJ = e.getGradDetJ(k);
            // Gauss points and determinant
            double w = gauss.getW(k);
            double detJ = e.getDetJ(k);
            for (int m = 0; m < nDim; ++m)
            {
                for (size_t i = 0; i < e.nodes.size(); ++i)
                {
                    size_t idx = i * nDim + m;
                    A1.col(idx) += w * mu * rhoU * (iJ * dSf).transpose() * (-iJ * dJ[idx] * dPhi) * detJ; // mu * rhoU * dgrad_phi * grad_psi * detj
                    A1.col(idx) += w * mu * rhoU * (-iJ * dJ[idx] * iJ * dSf).transpose() * dPhi * detJ;   // mu * rhoU * grad_phi * dgrad_psi * detj
                    A1.col(idx) += w * mu * rhoU * (iJ * dSf).transpose() * dPhi * dDetJ[idx];             // mu * rhoU * grad_phi * grad_psi * ddetj
                }
                for (size_t i = 0; i < eU.nodes.size(); ++i)
                {
                    size_t idx = i * nDim + m;
                    A2.col(idx) += w * mu * dRhoU * dPhiU.transpose() * (-iJU * dJU[idx] * dPhiU) * (iJ * dSf).transpose() * dPhi * detJ; // mu * drhoU * grad_phi * grad_psi * detj
                }
            }
            // Gradient of swithcing function, (rhoU-rho) * dmu * grad_phi * grad_psi * detj OR (rhoU-rho) * dmuU * grad_phi * grad_psi * detj
            if (machC >= machU)
            {
                double dM = fluid.mach->evalGrad(e, phi, k);
                for (int m = 0; m < nDim; ++m)
                    for (size_t i = 0; i < e.nodes.size(); ++i)
                        A1.col(i * nDim + m) += w * (-rho + rhoU) * dmu * dM * dPhi.transpose() * (-iJ * dJ[i * nDim + m] * dPhi) * (iJ * dSf).transpose() * dPhi * detJ;
            }
            else
            {
                double dM = fluid.mach->evalGrad(eU, phi, k);
                for (int m = 0; m < nDim; ++m)
                    for (size_t i = 0; i < eU.nodes.size(); ++i)
                        A2.col(i * nDim + m) += w * (-rho + rhoU) * dmu * dM * dPhiU.transpose() * (-iJU * dJU[i * nDim + m] * dPhiU) * (iJ * dSf).transpose() * dPhi * detJ;
            }
        }
    }
    return std::make_tuple(A1, A2);
}
