#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the nonlinear flow around rectangular 3D rae2822
# Adrien Crovato
#
# Test the direct and adjoint solvers, and the mesh morpher in 3D
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import dart.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors
import numpy as np

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 0
    M_inf = 0.8
    dim = 3

    # define dimension and mesh size
    spn = 1.0 # wing span
    lgt = 6.0 # channel length
    hgt = 6.0 # channel height
    wdt = 3.0 # channel width
    c_ref = 1.0 # reference length
    S_ref = spn # reference area
    fms = 1.0 # farfield mesh size
    nms = 0.02 # nearfield mesh size

    # parameters for mesh deformation
    alfa = 1*np.pi/180
    xc = 0.25
    dz_max = 0.2*spn

    # mesh the geometry and create mesh deformation handler
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms["msh"].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msN' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/rae_3.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    morpher = floD.morpher(msh, dim, 'wing')
    tms["msh"].stop()

    # set the problem and add fluid, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'wakeTip')
    solver = floD.newton(pbl) # create the solver now so element memory is up to date
    tms['pre'].stop()

    # deform the mesh (dummy rotation + bending)
    print(ccolors.ANSI_BLUE + 'PyDeforming...' + ccolors.ANSI_RESET)
    tms['deform'].start()
    morpher.savePos()
    rMat = np.matrix([[np.cos(alfa), np.sin(alfa)], [-np.sin(alfa), np.cos(alfa)]])
    vNods = {}
    for e in bnd.groups[0].tag.elems:
        for n in e.nodes:
            if n.no not in vNods:
                vNods[n.no] = n
    for no in vNods:
        nod = np.matrix([[vNods[no].pos[0]-xc], [vNods[no].pos[2]]])
        new = rMat * nod
        vNods[no].pos[0] = new[0,0]
        vNods[no].pos[2] = new[1,0] + dz_max/(spn*spn)*(vNods[no].pos[1]*vNods[no].pos[1])
    # deform the mesh
    morpher.deform()
    gmshWriter.save(msh.name+"_def")
    tms['deform'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    status = solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # solve adjoint problem
    print(ccolors.ANSI_BLUE + 'PyGradient...' + ccolors.ANSI_RESET)
    tms['adjoint'].start()
    adjoint = floD.adjoint(solver, morpher)
    adjoint.run()
    adjoint.save(gmshWriter)
    tms['adjoint'].stop()

    # compute norm of gradient wrt to mesh
    dClX = 0
    dCdX = 0
    for n in msh.nodes:
        dClX += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]))
        dCdX += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]))
    dClX = np.sqrt(dClX)
    dCdX = np.sqrt(dCdX)
    # recover gradient wrt to AoA from mesh gradient
    drot = np.array([[-np.sin(alfa), 0, np.cos(alfa)], [0, 0, 0], [-np.cos(alfa), 0, -np.sin(alfa)]])
    dClAoA = 0
    dCdAoA = 0
    for n in bnd.nodes:
        dx = drot.dot(np.array([n.pos[0] - 0.25, n.pos[1], n.pos[2]]))
        dClAoA += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(dx)
        dCdAoA += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(dx)

    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('       M    alpha       Cl       Cd   dCl_da   dCd_da')
    print('{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f} {5:8.4f}'.format(M_inf, alfa*180/np.pi, solver.Cl, solver.Cd, adjoint.dClAoa, adjoint.dCdAoa))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    if status > 0:
        raise Exception(ccolors.ANSI_RED + 'Solver failed to converge!' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('iteration count', solver.nIt, 7, 1, forceabs=True))
    tests.add(CTest('CL', solver.Cl, 0.21, 5e-2))
    tests.add(CTest('CD', solver.Cd, 0.008, 1e-2))
    tests.add(CTest('CS', solver.Cs, -0.020, 5e-2))
    tests.add(CTest('CM', solver.Cm, -0.080, 5e-2))
    tests.add(CTest('dCl/dAoA', adjoint.dClAoa, 3.0, 5e-2))
    tests.add(CTest('dCd/dAoA', adjoint.dCdAoa, 0.2, 5e-2))
    tests.add(CTest('dCl_dMsh', dClX, 2.2, 5e-2))
    tests.add(CTest('dCd_dMsh', dCdX, 0.2, 5e-2))
    tests.add(CTest('dCl/dAoA (msh)', dClAoA, adjoint.dClAoa, 5e-2))
    tests.add(CTest('dCd/dAoA (msh)', dCdAoA, adjoint.dCdAoa, 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
